/*  65816EMU.CPP - Emulates one mode of the 65816.  There are four different 
	versions depending on the mode of the processor.
*/

#undef XYTYPE
#undef ATYPE
#undef SIGNED_XYTYPE
#undef SIGNED_ATYPE
#undef XY_IS16BIT
#undef A_IS16BIT
#undef INDEX_X
#undef INDEX_Y
#undef ACCUM
#undef NV_SHIFT
#undef SIGN_BIT
#undef SIGN_BIT_XY
#undef SIGN_BIT_NUMBER
#undef BCDADDADJUST
#undef BCDSUBADJUST

#ifdef XY_16BIT
	#define XYTYPE unsigned short
	#define SIGNED_XYTYPE signed short
	#define XY_IS16BIT 1
	#define INDEX_X reg.X
	#define INDEX_Y reg.Y
#else
	#define XYTYPE unsigned char
	#define SIGNED_XYTYPE signed char
	#define XY_IS16BIT 0
	#define INDEX_X reg.X_lo
	#define INDEX_Y reg.Y_lo
#endif

#ifdef A_16BIT
	#define ATYPE unsigned short
	#define SIGNED_ATYPE signed short
	#define A_IS16BIT 1
	#define NV_SHIFT 8
	#define BCDADDADJUST bcdaddadjust_word
	#define BCDSUBADJUST bcdsubadjust_word
	#define ACCUM reg.A
#else
	#define ATYPE unsigned char
	#define SIGNED_ATYPE signed char
	#define A_IS16BIT 0
	#define NV_SHIFT 0
	#define BCDADDADJUST bcdaddadjust_byte
	#define BCDSUBADJUST bcdsubadjust_byte
	#define ACCUM reg.A_lo
#endif

#ifndef ADDRESSING_MODES
#define ADDRESSING_MODES

#define OPBYTE ((byte) opdata)
#define OPWORD ((word) opdata)
#define OPLONG opdata

	// Using these returns a pointer to the appropriate memory address.
	// The PC relative modes aren't listed since what is contained at the new location
	//   is irrelevant to the current instruction.
#define IMM_BYTE OPBYTE
	// OPC %#b
#define IMM_ACCUM ((ATYPE) opdata)
	// OPA %#
#define IMM_INDEX ((XYTYPE) opdata)
	// OPX %#i
#define ABS (SNESMEM ((DBR << 16) + OPWORD))
	// OPC %w
#define ABS_LONG (SNESMEM (OPLONG))
	// OPC $%L
#define DIR (SNESMEM ((D + OPBYTE) & 0xFFFF))
	// OPC <%b
#define DIR_INDIR_INDEX_Y ( _tmp = *(word*)SNESMEM(D + OPBYTE) + (DBR << 16) + INDEX_Y, SNESMEM(_tmp) )
	// OPC (<%b),Y
#define DIR_INDIR_LONG ( _tmp = (*(dword*)SNESMEM( (D + OPBYTE)&0xFFFF )) & 0xFFFFFF, SNESMEM (_tmp) )
	// OPC [<%b]
#define DIR_INDIR_LONG_INDEX_Y ( _tmp = ((*(dword*)SNESMEM( (D+OPBYTE)&0xFFFF )) & 0xFFFFFF) + INDEX_Y, SNESMEM (_tmp) )
	// OPC [<%b],Y
#define DIR_INDEX_INDIR_X ( _tmp = *(word *)SNESMEM( (((D+OPBYTE)&0xFFFF) + INDEX_X) & 0xFFFF ) + (DBR << 16), SNESMEM (_tmp) )
	// OPC (<%b,X)
#define DIR_INDEX_X ( SNESMEM(D + OPBYTE + INDEX_X) )
	// OPC <%b,X
#define DIR_INDEX_Y ( SNESMEM(D + OPBYTE + INDEX_Y) )
	// OPC <%b,Y
#define ABS_INDEX_X ( _tmp = (DBR << 16) + OPWORD + INDEX_X, SNESMEM(_tmp) )
	// OPC %w,X
#define ABS_INDEX_Y ( _tmp = (DBR << 16) + OPWORD + INDEX_Y, SNESMEM(_tmp) )
	// OPC %w, Y
#define ABS_LONG_INDEX_X ( SNESMEM(OPLONG + INDEX_X) )
	// OPC %L,X
#define ABS_INDIR ( /*_tmp = *(word*)SNESMEM(OPWORD),*/ SNESMEM (OPWORD) )
	// JML (%w) or JMP (%w)
#define DIR_INDIR ( _tmp = *(word*)SNESMEM((D + OPBYTE) & 0xFFFF) + (DBR << 16), SNESMEM (_tmp) )
	// OPC (<%b)
#define ABS_INDEX_INDIR_X ( _tmp = OPWORD + INDEX_X + (DBR << 16), SNESMEM(_tmp) )
	// OPC (%w,X)
#define ABS_INDEX_INDIR_X_JMP ( _tmp = OPWORD + INDEX_X + (PC & 0xFF0000), *(word*)SNESMEM(_tmp) | (PC & 0xFF0000) )
	// JMP (%w,X) - returns the new value for the PC.  As evidenced by Tetris Attack, the PBR (not DBR)
	// should be used for both bank values.
#define S_REL (SNESMEM (S + OPBYTE))
	// OPC <%b,S
#define S_REL_INDIR_INDEX_Y ( _tmp = *(word*)SNESMEM (S + OPBYTE) + (DBR << 16) + INDEX_Y, SNESMEM (_tmp) )
	// OPC (<%b,S),Y

#define PUSHBYTE(b) ( *(byte*)SNESMEM(S) = (byte)(b), (*(word*)&S)-- )
#define PUSHWORD(w) ( *(word*)SNESMEM(S-1) = (word)(w), (*(word*)&S) -= 2 )

#define PULLBYTE(db) ( (*(word*)&S)++, (db) = *(byte*)SNESMEM(S) )
#define PULLWORD(dw) ( (*(word*)&S) += 2, (dw) = *(word*)SNESMEM(S-1) )

#endif

#ifdef A_16BIT
	#define SIGN_BIT 0x8000
	#define SIGN_BIT_NUMBER 15
#else
	#define SIGN_BIT 0x80
	#define SIGN_BIT_NUMBER 7
#endif
#ifdef XY_16BIT
	#define SIGN_BIT_XY 0x8000
#else
	#define SIGN_BIT_XY 0x80
#endif

#define SAVE(A) (oldA = (A))
#define ADCSETCARRY(_tmpc) ((_tmpc) & (SIGN_BIT << 1) ? P |= CARRY : P &= ~CARRY)
#define SBCSETCARRY(_tmpc) ((_tmpc) & (SIGN_BIT << 1) ? P &= ~CARRY : P |= CARRY)
#define SETZERO(A) ((A) == 0 ? P |= ZERO : P &= ~ZERO)
#define SETNEGATIVE(A) ((A) & SIGN_BIT ? P |= NEGATIVE : P &= ~NEGATIVE)
#define SETNEGATIVEXY(XY) (((XY) & SIGN_BIT_XY) ? P |= NEGATIVE : P &= ~NEGATIVE)
#define SETOVERFLOW(A,oldA) ( ((A ^ oldA) & SIGN_BIT) && ((A>oldA?A-oldA:oldA-A) <= SIGN_BIT) ? P |= OVERFLOW : P &= ~OVERFLOW)
//#define SETOVERFLOW(A,oldA) ( (((SIGNED_ATYPE)A>(SIGNED_ATYPE)oldA?(int)((SIGNED_ATYPE)A)-(int)((SIGNED_ATYPE)oldA):(int)((SIGNED_ATYPE)oldA)-(int)((SIGNED_ATYPE)A))\
	> SIGN_BIT) ? P |= OVERFLOW : P &= ~OVERFLOW)
	// For an overflow condition:
	// - The signs have to change, and
	// - The value of (greater - smaller) is <= 32768

#define TRAPREAD(ptr)  (ptr >= registers && ptr <= (registers+0x4000) ? trapregread(ptr,ISWORDOP) : ptr)
#define TRAPWRITE(ptr) if (ptr >= registers && ptr <= (registers+0x4000)) trapregwrite(ptr,ISWORDOP)
	// TRAPWRITE should be called after the write and before any flags get set based on the new value.

#define ISWORDOP A_IS16BIT

#define P reg.P
#define S reg.S
#define D reg.D
#define E reg.E
#define DBR reg.DBR
#define PC reg.PC
#define ACCUM_16BIT reg.A
#define INDEX_X_16BIT reg.X
#define INDEX_Y_16BIT reg.Y
#define cycles scan_cycles

#define DOBREAK break

void EMUROUTINE ()
{
	//register dword PC;
	//static dword S;
	//static XYTYPE XYTYPE_X;
	//static XYTYPE XYTYPE_Y;
	//static byte DBR;
	//static word D;
	//static ATYPE ATYPE_A;
	//static byte P;
	//static boolean E; // Emulation mode flag (mostly ignored)
	static byte opcode;
	static dword opdata;
	dword _tmp, _tmp2, _tmpc;
	ATYPE oldA; // used by macros
	XYTYPE oldXY;
	ATYPE *amem;
	byte *mem;
	char unasm [20], lastop;

	(void)unasm;
	//PC = reg.PC;
	//S = reg.S;
	//XYTYPE_X = (XYTYPE) reg.X;
	//XYTYPE_Y = (XYTYPE) reg.Y;
	//DBR = reg.DBR;
	//D = reg.D;
	//ATYPE_A = (ATYPE) reg.A;
	//P = reg.P;
	//E = reg.E;

	//#define MAKE2(y) #y
	//#define MAKESTR(x) MAKE2(x)
	//if (debug_slice < 0) debug0 ("Entered: %s PC=%X P=%X S=%X", MAKESTR(EMUROUTINE), PC, P, S);
	assert ((S & 0xFFFF) == S);
	assert ((PC & 0xFFFFFF) == PC);
	while ((cycles & 0x80000000) == 0) {
		//if (!(PC & 0x8000)) {
			//addmessage ("warning! PC set to 0x%X", PC);
			//returntogui = true;
			//frame_cycles -= 100000;
			//cycles -= 200;
			//return;
		//}
		debug_instr++;

		opdata = *(dword*) SNESMEM(PC);
		//if (debug_instr >= 20250 && debug_instr < 20270) {
		//	unassemble (opdata,unasm,P);
		//	debug0 ("<-instr PC$%X (cycles=%d) asm:%s",PC,cycles,unasm);
		//}
//		debug thing
/*		debug_thing -= 2;
		*(word*)(expram + (debug_thing*6)+2) = (word)PC;
		*(word*)(expram + (debug_thing*6)+4) = (word)opdata;
		*(word*)(expram + (debug_thing*6)+6) = (word)S;
		*(word*)(expram + (debug_thing*6)+8) = (word)0;
		*(word*)(expram + (debug_thing*6)+10)= (word)E;
		*(word*)(expram + (debug_thing*6)+12)= (word)P;
		*(word*)expram = (0x6002 + debug_thing*6);
*/
		opcode = (byte)opdata;
		opdata >>= 8;

		// Another debug thing
		//if (opcode == 0x50 || opcode == 0x70)
			instr_tally[lastop][(P >> 3) & 1]++;
		lastop = opcode;
		
		switch (opcode) {
		case 0x00: //   "BRK", 2, 8, Has a signature word: %w two bytes ignored
			PC += 2; // Add 2 because RTI adds one more
			//PUSHBYTE (PC >> 16);
			//PUSHWORD ((word)PC);
			//PUSHBYTE (P);
			//PC = *(word *)SNESMEM (E? 0x00FFFE : 0x00FFE6);
			addmessage ("Warning: BRK encountered at $%LX", PC-2);
													cycles -= 200; frame_cycles -= 100000; returntogui = true;
			cycles -= 8;
			return;

			//DOBREAK;
		case 0x01: //   "ORA (<%b,X)", 2, 6,
			mem = DIR_INDEX_INDIR_X;
			ACCUM |= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x02: //   "CSP", 2, 8, Has one-byte signature %b
			PC += 2; // Add one because RTI adds one more -- wait, add 2??
			PUSHBYTE (PC >> 16);
			PUSHWORD ((word)PC);
			PUSHBYTE (P);
			PC = *(word *)SNESMEM (E? 0x00FFF4 : 0x00FFE4);
			cycles -= 8; DOBREAK;
		case 0x03: //   "ORA <%b,S", 2, 4,
			mem = S_REL;
			ACCUM |= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x04: //   "TSB <%b", 2, 5,
			amem = (ATYPE*)DIR;
			SETZERO (*amem & ACCUM);
			*amem |= ACCUM;
			TRAPWRITE((byte*)amem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x05: //   "ORA <%b", 2, 3,
			mem = DIR;
			ACCUM |= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x06: //   "SHL <%b", 2, 5,
			if (*(amem = (ATYPE*)DIR) & SIGN_BIT) P |= CARRY;
			else P &= ~CARRY;
			*amem <<= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x07: //   "ORA [<%b]", 2, 6,
			mem = DIR_INDIR_LONG;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x08: //   "PSH P", 1, 3, // PHP
			PUSHBYTE (P);
			PC += 1; cycles -= 3; DOBREAK;
		case 0x09: //   "ORA %#", 2, 2,
			ACCUM |= IMM_ACCUM;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0x0A: //   "SHL A", 1, 2,
			if (ACCUM & SIGN_BIT) P |= CARRY;
			else P &= ~CARRY;
			ACCUM <<= 1;
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x0B: //   "PSH D", 1, 4, // PHD Push Direct
			PUSHWORD (D);
			PC += 1; cycles -= 4; DOBREAK;
		case 0x0C: //   "TSB %w", 3, 6, 
			amem = (ATYPE*)ABS;
			SETZERO (*amem & ACCUM);
			*amem |= ACCUM;
			TRAPWRITE ((byte*)amem);
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x0D: //   "ORA %w", 3, 4,
			mem = ABS;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x0E: //   "SHL %w", 3, 6,
			if (*(amem = (ATYPE*)ABS) & SIGN_BIT) P |= CARRY;
			else P &= ~CARRY;
			*amem <<= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x0F: //   "ORA %L", 4, 5,
			mem = ABS_LONG;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x10: //   "BPL @%b", 2, 2,
			if (P & NEGATIVE) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0x11: //   "ORA (<%b),Y", 2, 5,
			mem = DIR_INDIR_INDEX_Y;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x12: //   "ORA (<%b)", 2, 5,
			mem = DIR_INDIR;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x13: //   "ORA (<%b,S),Y", 2, 7,
			mem = S_REL_INDIR_INDEX_Y;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x14: //   "TRB <%b", 2, 5,
			amem = (ATYPE*)DIR;
			SETZERO (*amem & ACCUM);
			*amem &= ~ACCUM;
			TRAPWRITE ((byte*)amem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x15: //   "ORA <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x16: //   "SHL <%b,X",2, 6,
			if (*(amem = (ATYPE*)DIR_INDEX_X) & SIGN_BIT) P |= CARRY;
			else P &= ~CARRY;
			*amem <<= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x17: //   "ORA [<%b],Y", 2, 6,
			mem = DIR_INDIR_LONG_INDEX_Y;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x18: //   "CLC", 1, 2,
			P &= ~CARRY;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x19: //   "ORA %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x1A: //   "INC A", 1, 2,
			ACCUM++;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x1B: //   "TAS", 1, 2,
			//(*((ATYPE*)&reg.A)) = ACCUM;
			S = ACCUM_16BIT;
			// No effect on flags
			PC += 1; cycles -= 2; DOBREAK;
		case 0x1C: //   "TRB %w", 3, 6,
			amem = (ATYPE*)ABS;
			SETZERO (*amem & ACCUM);
			*amem &= ~ACCUM;
			TRAPWRITE ((byte*)amem);
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x1D: //   "ORA %w,X", 3, 4,
			mem = ABS_INDEX_X;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x1E: //   "SHL %w,X", 3, 7,
			if (*(amem = (ATYPE*)ABS_INDEX_X) & SIGN_BIT) P |= CARRY;
			else P &= ~CARRY;
			*amem <<= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x1F: //   "ORA %L,X", 4, 5,
			mem = ABS_LONG_INDEX_X;
			ACCUM |= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x20: //   "JSR %w", 3, 6,
			PUSHWORD ((word)(PC + 2)); // TEST
			PC = (PC & 0xFFFF0000) | OPWORD;
			cycles -= 6; DOBREAK;
		case 0x21: //   "AND (<%b,X)", 2, 6,
			mem = DIR_INDEX_INDIR_X;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x22: //   "JSL %L", 4, 8,
			PUSHBYTE ((PC+3) >> 16);
			PUSHWORD ((word)PC + 3 ); // 3??
			PC = OPLONG;
			cycles -= 8; DOBREAK;
		case 0x23: //   "AND <%b,s", 2, 4,
			mem = S_REL;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x24: //   "BIT <%b", 2, 3,
			mem = DIR;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & ACCUM)
				P &= ~ZERO; else P |= ZERO;
			P = (P & ~(NEGATIVE | OVERFLOW)) | ((byte)(*amem >> NV_SHIFT) & (NEGATIVE | OVERFLOW));
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x25: //   "AND <%b", 2, 3,
			mem = DIR;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x26: //   "ROL <%b", 2, 5,
			amem = (ATYPE*)DIR;
			_tmp = *amem & SIGN_BIT;
			*amem = (ATYPE)((*amem << 1) | (P&1));
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 2; cycles -= 5; DOBREAK;
		case 0x27: //   "AND [<%b]", 2, 6,
			mem = DIR_INDIR_LONG;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x28: //   "PUL P", 1, 4, // PLP
			PC += 1; cycles -= 4;
			_tmp2 = P;
			PULLBYTE (P);
			if ((P & (MEMORY | INDEX)) != (_tmp2 & (MEMORY | INDEX))) {
				if (P & INDEX) { // if indexes are now 8-bit
					INDEX_X_16BIT &= 0xFF; // clear high byte
					INDEX_Y_16BIT &= 0xFF;
				}	
				goto saveregs_andexit;
			}
			DOBREAK;
		case 0x29: //   "AND %#", 2, 2,
			ACCUM &= IMM_ACCUM;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0x2A: //   "ROL A", 1, 2,
			//debug0 ("ROLA: Accum&SIGN_BIT %d signbit %d", ACCUM & SIGN_BIT, SIGN_BIT);
			//ACCUM = (ATYPE)opdata;
			_tmp = ACCUM & SIGN_BIT;
			ACCUM = (ATYPE)((ACCUM << 1) | (P&1));
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x2B: //   "PUL D", 1, 5, // PLD Pull Direct
			PULLWORD (D);
			SETZERO (D);
			if (D & 0x8000) P |= NEGATIVE;
			else P &= ~NEGATIVE; // Don't use SETNEGATIVE
			PC += 1; cycles -= 5; DOBREAK;
		case 0x2C: //   "BIT %w", 3, 4,
			mem = ABS;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & ACCUM)
				P &= ~ZERO; else P |= ZERO;
			P = (P & ~(NEGATIVE | OVERFLOW)) | ((byte)(*amem >> NV_SHIFT) & (NEGATIVE | OVERFLOW));
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x2D: //   "AND %w", 3, 4,
			mem = ABS;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x2E: //   "ROL %w", 3, 6,
			amem = (ATYPE*)ABS;
			_tmp = *amem & SIGN_BIT;
			*amem = (ATYPE)((*amem << 1) | (P&1));
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x2F: //   "AND %L", 4, 5,
			mem = ABS_LONG;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x30: //   "BMI @%b", 2, 2,
			if (!(P & NEGATIVE)) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0x31: //   "AND (<%b),Y", 2, 5,
			mem = DIR_INDIR_INDEX_Y;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x32: //   "AND (<%b)", 2, 5,
			mem = DIR_INDIR;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x33: //   "AND (<%b,S),Y", 2, 7,
			mem = S_REL_INDIR_INDEX_Y;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x34: //   "BIT <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & ACCUM)
				P &= ~ZERO; else P |= ZERO;
			P = (P & ~(NEGATIVE | OVERFLOW)) | ((byte)(*amem >> NV_SHIFT) & (NEGATIVE | OVERFLOW));
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x35: //   "AND <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x36: //   "ROL <%b,X", 2, 6,
			amem = (ATYPE*)DIR_INDEX_X;
			_tmp = *amem & SIGN_BIT;
			*amem = (ATYPE)((*amem << 1) | (P&1));
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x37: //   "AND [<%b],Y", 2, 6,
			mem = DIR_INDIR_LONG_INDEX_Y;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x38: //   "SEC", 1, 2,
			P |= CARRY;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x39: //   "AND %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x3A: //   "DEC A", 1, 2,
			ACCUM--;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x3B: //   "TSA", 1, 2,
			/*ACCUM = (ATYPE)*/ (ACCUM_16BIT = (word)S);
			SETZERO (ACCUM); // Flags should change
			SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x3C: //   "BIT %w,X", 3, 4,
			mem = ABS_INDEX_X;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & ACCUM)
				P &= ~ZERO; else P |= ZERO;
			P = (P & ~(NEGATIVE | OVERFLOW)) | ((byte)(*amem >> NV_SHIFT) & (NEGATIVE | OVERFLOW));
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x3D: //   "AND %w,X", 3, 4,
			mem = ABS_INDEX_X;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x3E: //   "ROL %w,X", 3, 7,
			mem = ABS_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmp = *amem & SIGN_BIT;
			*amem = (ATYPE)((*amem << 1) | (P&1));
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x3F: //   "AND %L,X", 4, 5,
			mem = ABS_LONG_INDEX_X;
			ACCUM &= *(ATYPE *)TRAPREAD(mem);
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x40: //   "RTI", 1, 7,
			_tmp = P;
			PULLBYTE (P);
			PULLWORD (PC);
			PULLBYTE (_tmp2);
			PC |= (_tmp2 << 16);
			cycles -= 7;
			if ((P & (MEMORY | INDEX)) != (_tmp & (MEMORY | INDEX))) {
				if (P & INDEX) { // if indexes are now 8-bit
					INDEX_X_16BIT &= 0xFF; // clear high byte
					INDEX_Y_16BIT &= 0xFF;
				}	
				goto saveregs_andexit;
			}
			DOBREAK;
		case 0x41: //   "XOR (<%b,X)", 2, 6,
			mem = DIR_INDEX_INDIR_X;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x42: //   "???", 2, 2,
			addmessage ("ERROR: INVALID OPCODE $42(WDM) AT $%LX", PC);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x43: //   "XOR <%b,S", 2, 4,
			mem = S_REL;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x44: //   "MVP XYA", 3, 7,
			cycles -= 7;
			DBR = OPBYTE;
			_tmp = DBR << 16;
			_tmp2 = (opdata << 8) & 0x00FF0000;
			*(byte*)SNESMEM (_tmp + INDEX_Y) = *(byte*)SNESMEM (_tmp2 + INDEX_X);
			INDEX_Y--;
			INDEX_X--;
			ACCUM--;
			if (ACCUM == (ATYPE) -1)
				PC += 3;
			DOBREAK;
		case 0x45: //   "XOR <%b", 2, 3,
			mem = DIR;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x46: //   "SHR <%b", 2, 5,
			if (*(amem = (ATYPE*)DIR) & 1) P |= CARRY;
			else P &= ~CARRY;
			*amem >>= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			P &= ~NEGATIVE;
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x47: //   "XOR [<%b]", 2, 6,
			mem = DIR_INDIR_LONG;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x48: //   "PSH A", 1, 3, // PHA
			#ifdef A_16BIT
			PUSHWORD (ACCUM);
			#else
			PUSHBYTE (ACCUM);
			#endif
			PC += 1; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x49: //   "XOR %#", 2, 2,
			ACCUM ^= IMM_ACCUM;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0x4A: //   "SHR A", 1, 2,
			if (ACCUM & 1) P |= CARRY;
			else P &= ~CARRY;
			ACCUM >>= 1;
			SETZERO (ACCUM);
			P &= ~NEGATIVE;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x4B: //   "PSH PBR", 1, 3, // PHK
			PUSHBYTE (PC >> 16);
			PC += 1; cycles -= 3; DOBREAK;
		case 0x4C: //   "JMP %w", 3, 3,
			PC = (PC & 0xFF0000) | OPWORD;
			cycles -= 3; DOBREAK;
		case 0x4D: //   "XOR %w", 3, 4,
			mem = ABS;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x4E: //   "SHR %w", 3, 6,
			if (*(amem = (ATYPE*)ABS) & 1) P |= CARRY;
			else P &= ~CARRY;
			*amem >>= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			P &= ~NEGATIVE;
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x4F: //   "XOR %L", 4, 5,
			mem = ABS_LONG;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x50: //   "BVC @%b", 2, 2,
			if (P & OVERFLOW) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0x51: //   "XOR (<%b),Y", 2, 5,
			mem = DIR_INDIR_INDEX_Y;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x52: //   "XOR (<%b)", 2, 5,
			mem = DIR_INDIR;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x53: //   "XOR (<%b,S),Y", 2, 7,
			mem = S_REL_INDIR_INDEX_Y;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x54: //   "MVN %w", 3, 7,
			cycles -= 7;
			DBR = OPBYTE;
			_tmp = DBR << 16;
			_tmp2 = (opdata << 8) & 0x00FF0000;
			*(byte*)SNESMEM (_tmp + INDEX_Y) = *(byte*)SNESMEM (_tmp2 + INDEX_X);
			INDEX_Y++;
			INDEX_X++;
			ACCUM--;
			if (ACCUM == (ATYPE) -1)
				PC += 3;
			DOBREAK;
		case 0x55: //   "XOR <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x56: //   "SHR <%b,X", 2, 6,
			mem = DIR_INDEX_X;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & 1) P |= CARRY;
			else P &= ~CARRY;
			*amem >>= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			P &= ~NEGATIVE;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x57: //   "XOR [<%b],Y", 2, 6,
			mem = DIR_INDIR_LONG_INDEX_Y;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x58: //   "CLI", 1, 2,
			P &= ~INT_DISABLE;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x59: //   "XOR %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x5A: //   "PSH Y", 1, 3, // PHY
			#ifdef XY_16BIT
			PUSHWORD (INDEX_Y);
			#else
			PUSHBYTE (INDEX_Y);
			#endif
			PC += 1; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0x5B: //   "TAD", 1, 2,
			//(*((ATYPE*)&reg.A)) = ACCUM;
			D = ACCUM_16BIT;
			SETZERO (D); // Flags are changed
			SETNEGATIVE (D);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x5C: //   "JML %L", 4, 4,
			PC = OPLONG;
			cycles -= 4; DOBREAK;
		case 0x5D: //   "XOR %w,X", 3, 4,
			mem = ABS_INDEX_X;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x5E: //   "SHR %w,X", 3, 7,
			mem = ABS_INDEX_X;
			if (*(amem = (ATYPE*)TRAPREAD(mem)) & 1) P |= CARRY;
			else P &= ~CARRY;
			*amem >>= 1;
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			P &= ~NEGATIVE;
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x5F: //   "XOR %L,X", 4, 5,
			mem = ABS_LONG_INDEX_X;
			ACCUM ^= *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x60: //   "RTS", 1, 6,
			_tmp = PC & 0xFFFF0000;
			PULLWORD (PC);
			PC |= _tmp;
			PC++;
			cycles -= 6; DOBREAK;
		case 0x61: //   "ADC (<%b,X)", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDEX_INDIR_X;
			amem = (ATYPE*)TRAPREAD(mem);			
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x62: //   "PSH @%w", 3, 6, // PER
			PC += 3;
			PUSHWORD ((word)PC + OPWORD);
			cycles -= 6; DOBREAK;
		case 0x63: //   "ADC <%b,S", 2, 4,
			SAVE(ACCUM);
			mem = S_REL;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x64: //   "CLR <%b", 2, 3,
			*(amem = (ATYPE*)DIR) = 0;
			TRAPWRITE ((byte*)amem);
			PC += 2; cycles -= 3+A_IS16BIT+A_IS16BIT; DOBREAK;
		case 0x65: //   "ADC <%b", 2, 3,
			SAVE(ACCUM);
			mem = DIR;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x66: //   "ROR <%b", 2, 5,
			amem = (ATYPE*)DIR;
			_tmp = *amem & 1;
			*amem = (*amem >> 1) | ((P&CARRY) << SIGN_BIT_NUMBER);
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x67: //   "ADC [<%b]", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDIR_LONG;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x68: //   "PUL A", 1, 4, // PLA
			#ifdef A_16BIT
			PULLWORD (ACCUM);
			#else
			PULLBYTE (ACCUM);
			#endif
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x69: //   "ADC %#", 2, 2,
			SAVE(ACCUM);
			_tmpc = (dword)ACCUM + (P&1) + IMM_ACCUM;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0x6A: //   "ROR A", 1, 2,
			_tmp = ACCUM & 1;
			ACCUM = (ACCUM >> 1) | ((P&CARRY) << SIGN_BIT_NUMBER);
			SETZERO (ACCUM);
			SETNEGATIVE (ACCUM);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x6B: //   "RTL", 1, 6,
			PULLWORD (PC);
			PULLBYTE (_tmp2);
			PC |= (_tmp2 << 16);
			PC++; // If JSL %L only incs PC by 3, here there must be a final inc.
			cycles -= 6; DOBREAK;
		case 0x6C: //   "JMP (%w)", 3, 5,
			PC = (PC & 0xFF0000) | *(word*)ABS_INDIR;
			cycles -= 5; DOBREAK;
		case 0x6D: //   "ADC %w", 3, 4,
			SAVE(ACCUM);
			mem = ABS;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x6E: //   "ROR %w", 3, 6, 
			amem = (ATYPE*)ABS;
			_tmp = *amem & 1;
			*amem = (*amem >> 1) | ((P&CARRY) << SIGN_BIT_NUMBER);
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x6F: //   "ADC %L", 4, 5,
			SAVE(ACCUM);
			mem = ABS_LONG;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x70: //   "BVS @%b", 2, 2,
			if (!(P & OVERFLOW)) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0x71: //   "ADC (<%b),Y", 2, 5,
			SAVE(ACCUM);
			mem = DIR_INDIR_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x72: //   "ADC (<%b)", 2, 5,
			SAVE(ACCUM);
			mem = DIR_INDIR;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x73: //   "ADC (<%b,s),Y", 2, 7,
			SAVE(ACCUM);
			mem = S_REL_INDIR_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x74: //   "CLR <%b,X", 2, 4,
			*(amem = (ATYPE*)DIR_INDEX_X) = 0;
			TRAPWRITE ((byte*)amem);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x75: //   "ADC <%b,X", 2, 4,
			SAVE(ACCUM);
			mem = DIR_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x76: //   "ROR <%b,X", 2, 6,
			amem = (ATYPE*)DIR_INDEX_X;
			_tmp = *amem & 1;
			*amem = (*amem >> 1) | ((P&CARRY) << SIGN_BIT_NUMBER);
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x77: //   "ADC [<%b],Y", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDIR_LONG_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x78: //   "SEI", 1, 2,
			P |= INT_DISABLE;
			PC += 1; cycles -= 2; DOBREAK;
		case 0x79: //   "ADC %w,Y", 3, 4,
			SAVE(ACCUM);
			mem = ABS_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x7A: //   "PUL Y", 1, 4, // PLY
			#ifdef XY_16BIT
			PULLWORD (INDEX_Y);
			#else
			PULLBYTE (INDEX_Y);
			#endif
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 1; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0x7B: //   "TDA", 1, 2,
			/*ACCUM = (ATYPE)*/ (ACCUM_16BIT = D);
			SETZERO (ACCUM); // Flags are changed
			SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x7C: //   "JMP (%w,X)", 3, 6,
			PC = ABS_INDEX_INDIR_X_JMP;
			cycles -= 6; DOBREAK;
		case 0x7D: //   "ADC %w,X", 3,4,
			SAVE(ACCUM);
			mem = ABS_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x7E: //   "ROR %w,X", 3,7,
			amem = (ATYPE*)ABS_INDEX_X;
			_tmp = *amem & 1;
			*amem = (*amem >> 1) | ((P&CARRY) << SIGN_BIT_NUMBER);
			TRAPWRITE ((byte*)amem);
			SETZERO (*amem);
			SETNEGATIVE (*amem);
			if (_tmp) P |= CARRY; else P &= ~CARRY;
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x7F: //   "ADC %L,X",4,5,
			SAVE(ACCUM);
			mem = ABS_LONG_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM + (P&1) + *amem;
			ADCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDADDADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x80: //   "BRA @%b", 2, 2,
			PC += 2 + ((signed char) opdata);
			cycles -= 2; DOBREAK;
		case 0x81: //   "STA (<%b,X)", 2, 6,
			*(ATYPE*)(mem = DIR_INDEX_INDIR_X) = ACCUM;
			TRAPWRITE (mem);			
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x82: //   "BRL @%w", 3, 3,
			_tmp = PC & 0xFF0000;
			PC = ((PC + 3 + (signed short) OPWORD) & 0xFFFF) | _tmp;
			cycles -= 3; DOBREAK;
		case 0x83: //   "STA <%b,S", 2, 4,
			*(ATYPE*)(mem = S_REL) = ACCUM;
			TRAPWRITE (mem);			
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x84: //   "STY <%b", 2, 3,
			*(XYTYPE*)(mem = DIR) = INDEX_Y;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0x85: //   "STA <%b", 2, 3,
			*(ATYPE*)(mem = DIR) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0x86: //   "STX <%b", 2, 3,
			*(XYTYPE*)(mem = DIR) = INDEX_X;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0x87: //   "STA [<%b]", 2, 6,
			*(ATYPE*)(mem = DIR_INDIR_LONG) = ACCUM;
			TRAPWRITE (mem);			
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x88: //   "DEY", 1, 2,
			INDEX_Y--;
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x89: //   "BIT %#", 2, 2,
			if (ACCUM & IMM_ACCUM)
				P &= ~ZERO; else P |= ZERO;
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0x8A: //   "TXA", 1, 2,
			//(*((XYTYPE*)&reg.X)) = INDEX_X;
			ACCUM = (ATYPE)INDEX_X_16BIT;
			//ACCUM = (ATYPE)INDEX_X;
			SETZERO (ACCUM); // Flags are changed
			SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x8B: //   "PSH DBR", 1, 3, // PHB
			PUSHBYTE (DBR);
			PC += 1; cycles -= 3; DOBREAK;
		case 0x8C: //   "STY %w", 3, 4,
			*(XYTYPE*)(mem = ABS) = INDEX_Y;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0x8D: //   "STA %w", 3, 4,
			*(ATYPE*)(mem = ABS) = ACCUM;
			TRAPWRITE (mem);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x8E: //   "STX %w", 3, 4,
			*(XYTYPE*)(mem = ABS) = INDEX_X;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0x8F: //   "STA %L", 4, 5,
			*(ATYPE*)(mem = ABS_LONG) = ACCUM;
			TRAPWRITE (mem);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x90: //   "BCC @%b", 2, 2,
			if (P & CARRY) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0x91: //   "STA (<%b),Y", 2, 6,
			*(ATYPE*)(mem = DIR_INDIR_INDEX_Y) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x92: //   "STA (<%b)", 2, 5,
			*(ATYPE*)(mem = DIR_INDIR) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x93: //   "STA (<%b,S),Y", 2, 7,
			*(ATYPE*)(mem = S_REL_INDIR_INDEX_Y) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0x94: //   "STY <%b,X", 2, 4,
			*(XYTYPE*)(mem = DIR_INDEX_X) = INDEX_Y;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 2; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0x95: //   "STA <%b,X", 2, 4,
			*(ATYPE*)(mem = DIR_INDEX_X) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x96: //   "STX <%b,Y", 2, 4,
			*(XYTYPE*)(mem = DIR_INDEX_Y) = INDEX_X;
				#undef ISWORDOP
				#define ISWORDOP XY_IS16BIT
			TRAPWRITE (mem);
				#undef ISWORDOP
				#define ISWORDOP A_IS16BIT		
			PC += 2; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0x97: //   "STA [<%b],Y", 2, 6,
			*(ATYPE*)(mem = DIR_INDIR_LONG_INDEX_Y) = ACCUM;
			TRAPWRITE (mem);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0x98: //   "TYA", 1, 2,
			//(*((XYTYPE*)&reg.Y)) = INDEX_Y;
			ACCUM = (ATYPE)INDEX_Y_16BIT;
			//ACCUM = (ATYPE)INDEX_Y;
			SETZERO (ACCUM); // Flags are changed
			SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x99: //   "STA %w,Y", 3, 5,
			*(ATYPE*)(mem = ABS_INDEX_Y) = ACCUM;
			TRAPWRITE (mem);
			PC += 3; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x9A: //   "TXS", 1, 2,
			S = INDEX_X_16BIT;
			// Has no effect on flags
			PC += 1; cycles -= 2; DOBREAK;
		case 0x9B: //   "TXY", 1, 2,
			INDEX_Y = INDEX_X;
			SETZERO (INDEX_Y); // Flags are changed
			SETNEGATIVEXY (INDEX_Y);
			PC += 1; cycles -= 2; DOBREAK;
		case 0x9C: //   "CLR %w", 3, 4,
			*(ATYPE*)(mem = ABS) = 0;
			TRAPWRITE (mem);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0x9D: //   "STA %w,X", 3, 5,
			*(ATYPE*)(mem = ABS_INDEX_X) = ACCUM;
			TRAPWRITE (mem);
			PC += 3; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x9E: //   "CLR %w,X", 3, 5,
			*(ATYPE*)(mem = ABS_INDEX_X) = 0;
			TRAPWRITE (mem);
			PC += 3; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0x9F: //   "STA %L,X", 4, 5,
			*(ATYPE*)(mem = ABS_LONG_INDEX_X) = ACCUM;
			TRAPWRITE (mem);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xA0: //   "LDY %#i", 2, 2,
			INDEX_Y = IMM_INDEX;
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 2+XY_IS16BIT; cycles -= 2+XY_IS16BIT; DOBREAK;
		case 0xA1: //   "LDA (<%b,X)", 2, 6,
			mem = DIR_INDEX_INDIR_X;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xA2: //   "LDX %#i", 2, 2,
			INDEX_X = IMM_INDEX;
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 2+XY_IS16BIT; cycles -= 2+XY_IS16BIT; DOBREAK;
		case 0xA3: //   "LDA <%b,S", 2, 4,
			mem = S_REL;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xA4: //   "LDY <%b", 2, 3,
			mem = DIR;
			INDEX_Y = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0xA5: //   "LDA <%b", 2, 3,
			mem = DIR;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0xA6: //   "LDX <%b", 2, 3,
			mem = DIR;
			INDEX_X = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0xA7: //   "LDA [<%b]", 2, 6,
			mem = DIR_INDIR_LONG;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xA8: //   "TAY", 1, 2,
			//(*((ATYPE*)&reg.A)) = ACCUM;
			INDEX_Y = (XYTYPE)ACCUM_16BIT;
			//INDEX_Y = (XYTYPE)ACCUM;
			SETZERO (INDEX_Y); // Flags are changed
			SETNEGATIVEXY (INDEX_Y);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xA9: //   "LDA %#", 2, 2,
			ACCUM = IMM_ACCUM;
			SETZERO (ACCUM); 
			/*debug0 ("LDA # - P=%X after setzero", P);
			if ((ACCUM) & SIGN_BIT)
				P |= NEGATIVE;
			else
				P &= ~NEGATIVE;
			//(((ACCUM) & SIGN_BIT) ?  : );
			debug0 ("(byte)0xB5 & ~NEGATIVE = %X  NEGATIVE=%X, ~=%X", (byte)0xB5 & ~NEGATIVE, NEGATIVE, ~(byte)NEGATIVE);*/
			SETNEGATIVE (ACCUM);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0xAA: //   "TAX", 1, 2,
			//(*((ATYPE*)&reg.A)) = ACCUM;
			INDEX_X = (XYTYPE)ACCUM_16BIT;
			//INDEX_X = (XYTYPE)ACCUM;
			SETZERO (INDEX_X); // Flags are changed
			SETNEGATIVEXY (INDEX_X);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xAB: //   "PUL DBR", 1, 4, // PLB
			PULLBYTE (DBR);
			SETZERO (DBR);
			if (DBR & 0x80) P |= NEGATIVE;
			else P &= ~NEGATIVE;
			PC += 1; cycles -= 4; DOBREAK;
		case 0xAC: //   "LDY %w", 3, 4,
			mem = ABS;
			INDEX_Y = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xAD: //   "LDA %w", 3, 4,
			mem = ABS;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xAE: //   "LDX %w", 3, 4,
			mem = ABS;
			INDEX_X = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xAF: //   "LDA %L", 4, 5,
			mem = ABS_LONG;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xB0: //   "BCS @%b", 2, 2,
			if (!(P & CARRY)) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0xB1: //   "LDA (<%b),Y", 2, 5,
			mem = DIR_INDIR_INDEX_Y;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xB2: //   "LDA (<%b)", 2, 5,
			mem = DIR_INDIR;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xB3: //   "LDA (<%b,S),Y", 2, 7,
			mem = S_REL_INDIR_INDEX_Y;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0xB4: //   "LDY <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			INDEX_Y = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 2; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xB5: //   "LDA <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xB6: //   "LDX <%b,Y", 2, 4,
			mem = DIR_INDEX_Y;
			INDEX_X = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 2; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xB7: //   "LDA [<%b],Y", 2, 6,
			mem = DIR_INDIR_LONG_INDEX_Y;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 2; cycles -= 6+A_IS16BIT;
			DOBREAK;
		case 0xB8: //   "CLV", 1, 2,
			P &= ~OVERFLOW;
			PC += 1; cycles -= 2; DOBREAK;
		case 0xB9: //   "LDA %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xBA: //   "TSX", 1, 2,
			INDEX_X = (XYTYPE) *(XYTYPE*)&S;
			SETZERO (INDEX_X); // Flags are changed
			SETNEGATIVEXY (INDEX_X);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xBB: //   "TYX", 1, 2,
			INDEX_X = INDEX_Y;
			SETZERO (INDEX_X); // Flags are changed
			SETNEGATIVE (INDEX_X);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xBC: //   "LDY %w,X", 3, 4,
			mem = ABS_INDEX_X;
			INDEX_Y = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xBD: //   "LDA %w,X", 3, 4,
			mem = ABS_INDEX_X;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xBE: //   "LDX %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			INDEX_X = *(XYTYPE*)TRAPREAD(mem);
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xBF: //   "LDA %L,X",4, 5,
			mem = ABS_LONG_INDEX_X;
			ACCUM = *(ATYPE*)TRAPREAD(mem);
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xC0: //   "CPY %#i", 2, 2,
			oldXY = INDEX_Y - IMM_INDEX;
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_Y)
				P &= ~CARRY; else P |= CARRY;
			PC += 2+XY_IS16BIT; cycles -= 2+XY_IS16BIT; DOBREAK;
		case 0xC1: //   "CMP (<%b,X)", 2, 6,
			mem = DIR_INDEX_INDIR_X;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xC2: //   "CLP %#b", 2, 3,
			PC += 2; cycles -= 3;
			if (((P & ~IMM_BYTE) & (MEMORY | INDEX)) != (P & (MEMORY | INDEX))) { // Change
				P &= ~IMM_BYTE;
				goto saveregs_andexit;
			}
			P &= ~IMM_BYTE;
			DOBREAK;
		case 0xC3: //   "CMP <%b,S", 2, 4,
			mem = S_REL;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xC4: //   "CPY <%b", 2, 3,
			mem = DIR;
			oldXY = INDEX_Y - *(XYTYPE*)TRAPREAD(mem);
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_Y)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0xC5: //   "CMP <%b", 2, 3,
			mem = DIR;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0xC6: //   "DEC <%b", 2, 5,
			mem = DIR;
			(*(amem = (ATYPE*)TRAPREAD(mem)))--;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xC7: //   "CMP [<%b]", 2,6,
			mem = DIR_INDIR_LONG;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xC8: //   "INY", 1, 2,
			INDEX_Y++;
			SETZERO (INDEX_Y); SETNEGATIVEXY (INDEX_Y);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xC9: //   "CMP %#", 2, 2,
			oldA = ACCUM - IMM_ACCUM;
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0xCA: //   "DEX", 1, 2,
			INDEX_X--;
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xCB: //   "WAI", 1, 3,
			if (state.end_wai && state.within_wai) {
				PC += 1;
				state.end_wai = false;
				state.within_wai = false;
			} else {
				state.within_wai = true;
				state.end_wai = false;
			}
			cycles = -3; DOBREAK;
		case 0xCC: //   "CPY %w", 3, 4,
			mem = ABS;
			oldXY = INDEX_Y - *(XYTYPE*)TRAPREAD(mem);
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_Y)
				P &= ~CARRY; else P |= CARRY;
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xCD: //   "CMP %w", 3, 4,
			mem = ABS;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xCE: //   "DEC %w", 3, 6,
			mem = ABS;
			(*(amem = (ATYPE*)TRAPREAD(mem)))--;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xCF: //   "CMP %L", 4, 5,
			mem = ABS_LONG;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xD0: //   "BNE @%b", 2, 2,
			if (P & ZERO) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0xD1: //   "CMP (<%b),Y", 2, 5,
			mem = DIR_INDIR_INDEX_Y;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xD2: //   "CMP (<%b)", 2, 5,
			mem = DIR_INDIR;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xD3: //   "CMP (<%b,S),Y", 2, 7,
			mem = S_REL_INDIR_INDEX_Y;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0xD4: //   "PSH <%b", 2, 6, // PEI
			PUSHWORD (*(word*)DIR);
			PC += 2; cycles -= 6; DOBREAK;
		case 0xD5: //   "CMP <%b,X", 2, 4,
			mem = DIR_INDEX_X;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xD6: //   "DEC <%b,X", 2, 6,
			mem = DIR_INDEX_X;
			(*(amem = (ATYPE*)TRAPREAD(mem)))--;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xD7: //   "CMP [<%b],Y", 2, 6,
			mem = DIR_INDIR_LONG_INDEX_Y;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xD8: //   "CLD", 1, 2,
			P &= ~BCD;
			PC += 1; cycles -= 2; DOBREAK;
		case 0xD9: //   "CMP %w,Y", 3, 4,
			mem = ABS_INDEX_Y;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xDA: //   "PSH X", 1, 3, // PHX
			#ifdef XY_16BIT
			PUSHWORD (INDEX_X);
			#else
			PUSHBYTE (INDEX_X);
			#endif
			PC += 1; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0xDB: //   "HLT", 1, 3, // STP
			addmessage ("Warning: SNES HALTED (Clock stopped)!");
			//PC += 1; cycles -= 3;
			cycles = -1;
			DOBREAK;
		case 0xDC: //   "JML (%w)", 3, 6,
			PC = *(long*)ABS_INDIR & 0xFFFFFF;
			cycles -= 6; DOBREAK;
		case 0xDD: //   "CMP %w,X", 3, 4,
			mem = ABS_INDEX_X;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xDE: //   "DEC %w,X", 3, 7,
			mem = ABS_INDEX_X;
			(*(amem = (ATYPE*)TRAPREAD(mem)))--;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0xDF: //   "CMP %L,X", 4, 5,
			mem = ABS_LONG_INDEX_X;
			oldA = ACCUM - *(ATYPE*)TRAPREAD(mem);
			SETNEGATIVE (oldA);	SETZERO (oldA);
			if (oldA > ACCUM)
				P &= ~CARRY; else P |= CARRY;
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xE0: //   "CPX %#i", 2, 2,
			oldXY = INDEX_X - IMM_INDEX;
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_X)
				P &= ~CARRY; else P |= CARRY;
			PC += 2+XY_IS16BIT; cycles -= 2+XY_IS16BIT; DOBREAK;
		case 0xE1: //   "SBC (<%b,X)", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDEX_INDIR_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xE2: //   "SEP %#b", 2, 3,
			PC += 2; cycles -= 3;
			if (((P | IMM_BYTE) & (MEMORY | INDEX)) != (P & (MEMORY | INDEX))) { // Change
				P |= IMM_BYTE;
				if (IMM_BYTE & INDEX) { // if switching from 16-bit to 8-bit indeXes,
					INDEX_X_16BIT &= 0xFF; // clear high byte
					INDEX_Y_16BIT &= 0xFF;
				}	
				goto saveregs_andexit;
			}
			P |= IMM_BYTE;
			DOBREAK;
		case 0xE3: //   "SBC <%b,S", 2, 4,
			SAVE(ACCUM);
			mem = S_REL;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xE4: //   "CPX <%b", 2, 3,
			mem = DIR;
			oldXY = INDEX_X - *(XYTYPE*)TRAPREAD(mem);
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_X)
				P &= ~CARRY; else P |= CARRY;
			PC += 2; cycles -= 3+XY_IS16BIT; DOBREAK;
		case 0xE5: //   "SBC <%b", 2, 3,
			SAVE(ACCUM);
			mem = DIR;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 3+A_IS16BIT; DOBREAK;
		case 0xE6: //   "INC <%b", 2, 5,
			mem = DIR;
			(*(amem = (ATYPE*)TRAPREAD(mem)))++;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xE7: //   "SBC [<%b]", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDIR_LONG;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xE8: //   "INX", 1, 2,
			INDEX_X++;
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 1; cycles -= 2; DOBREAK;
		case 0xE9: //   "SBC %#", 2, 2,
			SAVE(ACCUM);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + IMM_ACCUM);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2+A_IS16BIT; cycles -= 2+A_IS16BIT; DOBREAK;
		case 0xEA: //   "NOP", 1, 2,
			PC += 1; cycles -= 2; DOBREAK;
		case 0xEB: //   "SWA", 1, 3,
			//(*((ATYPE*)&reg.A)) = ACCUM;
			_tmp = ACCUM_16BIT & 0xFF;
			ACCUM_16BIT >>= 8;
			ACCUM_16BIT |= _tmp << 8;
			//ACCUM = (ATYPE) reg.A;
			SETZERO (ACCUM); SETNEGATIVE (ACCUM);
			PC += 1; cycles -= 3; DOBREAK;
		case 0xEC: //   "CPX %w", 3, 4,
			mem = ABS;
			oldXY = INDEX_X - *(XYTYPE*)TRAPREAD(mem);
			SETNEGATIVEXY (oldXY); SETZERO (oldXY);
			if (oldXY > INDEX_X)
				P &= ~CARRY; else P |= CARRY;
			PC += 3; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xED: //   "SBC %w", 3, 4,
			SAVE(ACCUM);
			mem = ABS;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xEE: //   "INC %w", 3, 6,
			mem = ABS;
			(*(amem = (ATYPE*)TRAPREAD(mem)))++;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 3; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xEF: //   "SBC %L", 4, 5,
			SAVE(ACCUM);
			mem = ABS_LONG;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xF0: //   "BEQ @%b", 2, 2,
			if (!(P & ZERO)) { // False
				PC += 2; cycles -= 2; DOBREAK;
			}
			PC += 2 + ((signed char) opdata);
			cycles -= 3; DOBREAK;
		case 0xF1: //   "SBC (<%b),Y", 2, 5,
			SAVE(ACCUM);
			mem = DIR_INDIR_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xF2: //   "SBC (<%b)", 2, 5,
			SAVE(ACCUM);
			mem = DIR_INDIR;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 5+A_IS16BIT; DOBREAK;
		case 0xF3: //   "SBC (<%b,S),Y", 2, 7,
			SAVE(ACCUM);
			mem = S_REL_INDIR_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0xF4: //   "PSH %#w", 3, 5, // PEA
			PUSHWORD (OPWORD);
			PC += 3; cycles -= 5; DOBREAK;
		case 0xF5: //   "SBC <%b,X", 2, 4,
			SAVE(ACCUM);
			mem = DIR_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xF6: //   "INC <%b,X", 2, 6,
			mem = DIR_INDEX_X;
			(*(amem = (ATYPE*)TRAPREAD(mem)))++;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xF7: //   "SBC [<%b],Y", 2, 6,
			SAVE(ACCUM);
			mem = DIR_INDIR_LONG_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 2; cycles -= 6+A_IS16BIT; DOBREAK;
		case 0xF8: //   "SED", 1, 2,
			P |= BCD;
			PC += 1; cycles -= 2; DOBREAK;
		case 0xF9: //   "SBC %w,Y", 3, 4,
			SAVE(ACCUM);
			mem = ABS_INDEX_Y;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xFA: //   "PUL X", 1, 4, // PLX
			#ifdef XY_16BIT
			PULLWORD (INDEX_X);
			#else
			PULLBYTE (INDEX_X);
			#endif
			SETZERO (INDEX_X); SETNEGATIVEXY (INDEX_X);
			PC += 1; cycles -= 4+XY_IS16BIT; DOBREAK;
		case 0xFB: //   "XCE", 1, 2,
			PC += 1; cycles -= 2;
			_tmp2 = P & CARRY;
			if (E) P |= CARRY; else P &= ~CARRY;
			E = _tmp2;
			if (E) {
				P |= INDEX | MEMORY;
				S = (S & 0xFF) | 0x100;
				INDEX_X_16BIT &= 0xFF;
				INDEX_Y_16BIT &= 0xFF;
				D = 0;
				goto saveregs_andexit;
			}
			DOBREAK;
		case 0xFC: //   "JSR (%w,X)", 3, 6,
			PUSHWORD ((word)(PC + 2)); // TEST
			PC = ABS_INDEX_INDIR_X_JMP;
			cycles -= 6; DOBREAK;
		case 0xFD: //   "SBC %w,X", 3, 4,
			SAVE(ACCUM);
			mem = ABS_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 3; cycles -= 4+A_IS16BIT; DOBREAK;
		case 0xFE: //   "INC %w,X", 3, 7,
			mem = ABS_INDEX_X;
			(*(amem = (ATYPE*)TRAPREAD(mem)))++;
			SETZERO (*amem); SETNEGATIVE (*amem);
			PC += 3; cycles -= 7+A_IS16BIT; DOBREAK;
		case 0xFF: //   "SBC %L,X", 4, 5
			SAVE(ACCUM);
			mem = ABS_LONG_INDEX_X;
			amem = (ATYPE*)TRAPREAD(mem);
			_tmpc = (dword)ACCUM - ((P&CARRY ? 0 : 1) + *amem);
			SBCSETCARRY(_tmpc);
			ACCUM = _tmpc;
			if (P & BCD) BCDSUBADJUST (ACCUM);
			SETZERO(ACCUM);
			SETNEGATIVE(ACCUM); SETOVERFLOW(ACCUM, oldA);
			PC += 4; cycles -= 5+A_IS16BIT; DOBREAK;
		}
	}

saveregs_andexit:
	//reg.PC = PC;
	//reg.P = P;
	//debug0 ("reg.P set to %X", reg.P);
	//(*((XYTYPE*)&reg.X)) = INDEX_X;
	//(*((XYTYPE*)&reg.Y)) = INDEX_Y;
	//(*((ATYPE*)&reg.A)) = ACCUM;
	//reg.S = S;
	//reg.D = D;
	//reg.DBR = DBR;
	//reg.E = E;
	//return cycles;
	return;
}

#undef DOBREAK

#undef P
#undef S
#undef D
#undef E
#undef DBR
#undef PC
#undef ACCUM_16BIT
#undef INDEX_X_16BIT
#undef INDEX_Y_16BIT
#undef cycles

